// nfsp iod = Input Output Daemon NG = a multithreaded I/O daemon
//
//   * 1 thread waits for UDP datagrams (read/write/ping) and fills buffers
//   (slots) that will be processed by the io_thr
//
//   * 1 thread waits for TCP remove messages and fills buffers (slots) that
//   will be processed by the io_thr
//
//   * 1 (or more) threads act(s) as I/O daemons handling disk operation and
//   send replies to the clients when needed
//
// This server waits for requests sent by the main server (rpc.nfspd) and
// generates an answer for a client (by spoofing it).
//
// Changelog: look into the CVS log !
//
// Contributors:
//   Adrien Lebre <adrien.lebre@imag.fr>
//
// Author: Pierre Lombard <pierre.lombard@imag.fr>
//
// Licence: GPLv2

#ifndef _REENTRANT
# error "_REENTRANT must be defined for pthread support"
#endif

#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <pwd.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <syslog.h>

// Included for PF_PACKET (with <sys/socket.h> just above)
#include <features.h>    /* for the glibc version number */
#if __GLIBC__ >= 2 && __GLIBC_MINOR >= 1
#  include <netpacket/packet.h>
#  include <net/ethernet.h>     /* the L2 protocols */
#else
#  include <asm/types.h>
#  include <linux/if_packet.h>
#  include <linux/if_ether.h>   /* The L2 protocols */
#endif

#include <sys/time.h>
#include <sys/resource.h>


#include <endian.h>
#if (__BYTE_ORDER != __LITTLE_ENDIAN)
# error "Parts of this code depends on CPU endianness: it was only tested on Linux i386"
#endif

#include "iodng.h"
#include "udpspoofng.h"	// used for UDP spoofing
#include "in_cksum.h"	// used for IP checksum

//#define SHOWINFO 1
//#include "dbg.h"
//
#include "logging.h"
#include "ipstring.h"

// FIFO_PROFILING will show verbose information on the level of filling of the buffers
//#define FIFO_PROFILING 1

// Should extrachecks be done - since they should have been done by the server
#define PARANOID 1

// How many messages should be spooled ?
#define MSG_ARRAY_SIZE 256

#define HAS_ROOT 0
#define HAS_PRIO 0

#if HAS_PRIO  && (! HAS_ROOT)
#error "priority tweaking requires root permissions"
#endif

#if HAS_ROOT && HAS_PRIO
// What should be the NICE adjustement for the threads ?
#define PRIO_UDP_THR	-15
#define PRIO_TCP_THR	-10
#define PRIO_IO_THR	10
#endif

// How many connections requests should be backlogged ?
#define TCP_BACKLOG		5

#define PROGNAME "iodng"
#define UDP_MAXLEN 65535

typedef enum {FREE, RECV, FILLED, USED} msg_state_t;

// The array in which messages are stored, waiting for being processed
// Note: you must hold msg_mutex to access these variables
static pthread_mutex_t msg_mutex = PTHREAD_MUTEX_INITIALIZER;
static msg_state_t     msg_state[MSG_ARRAY_SIZE];
static pthread_cond_t  msg_cond_free = PTHREAD_COND_INITIALIZER;
static int             msg_free = MSG_ARRAY_SIZE;
static int iIndexFREE = 0;	// First free slot
static int iIndexFILLED = 0;	// Last slot filled
static pthread_cond_t  msg_cond_filled = PTHREAD_COND_INITIALIZER;
static int             msg_filled = 0;

// This array may be written to once an indix has been found in msg_state and once it has been locked
static msg_iod_t       msg_array[MSG_ARRAY_SIZE];


// These 2 sockets are created  by the main thread before any
// fork()/pthread_create() call is done. After that, these variables
// are only read (though the corresponding socket may be written to)
#if HAS_ROOT
static int iprawsocket = -1;
static int llsocket = -1;
#endif
static int udpsocket = -1;
static int tcpsocket = -1;
static int forward_step1_read_udpsock = -1;	// Forward Step1 Read UDP socket
static int forward_step1_write_udpsock = -1;	// Forward Step1 Read UDP socket
#ifdef WRITEMIRROR_SUPPORT
static int write_mirror_udpsock = -1;		// Write Mirror UDP socket
#endif /* WRITEMIRROR_SUPPORT */

// Command line options (set to default values)
static int foreground = 0;
static int nioservers = 2;
static int listenport = 20000;
#if HAS_ROOT
static struct passwd *user = NULL;
#endif
static char *datapath = NULL;

// Mode used to send packets
typedef enum {XMITMODE_ARP = 0, XMITMODE_IP, XMITMODE_UDP, XMITMODE_LAST, } xmitmode_t;
static xmitmode_t xmitmode = XMITMODE_UDP;

// RPC headers don't have a constant size that's why the following struct has
// been defined. It just runs fine with NFS read or write replies with the
// Linux nfs usermode server and linux clients.
//
// Explanations : see rpc/rpc_msg.h
// 		u_long		rm_xid;						+4
//		enum msg_iod_type	rm_direction;				+4
//		struct reply_body RM_rmb {
//			enum reply_stat rp_stat;				+4
//			struct accepted_reply RP_ar {
//				struct opaque_auth ar_verf {
//				        enum_t  oa_flavor;			+4
//					caddr_t oa_base;			+4
//					[but is NULL since auth NULL => nothing more]
//		}
//		enum accept_stat ar_stat;					+4
//
// which does a 24-byte header defined below
struct myrpc_msg {
	u_long rm_xid;
	enum msg_type rm_direction;
	enum reply_stat rp_stat;
	enum_t oa_flavor;					// must be NULL
	caddr_t oa_base;					// must be NULL
	enum accept_stat ar_stat;
};


// This function implements a one byte hash based on the inode and the seed and is used
// to find the 1st level directory in which is stored the file
static inline unsigned char hash(unsigned int inode, unsigned int seed) {
	unsigned int hash;
	hash = inode ^ seed;
	hash ^= hash >> 16;
	return (0xFF & hash) % NB_IODS_SUBDIR;
}

// Parameters:
//  m		IN	message
//  filename	OUT	string reply must be at least KEYMAXSIZE
//
// This is the max length of a file associated to a key
// WARNING: if changed => do_purge()

// path/file looks like XX/iXXXXXXXXsXXXXXXXX
#define KEYMAXSIZE (2+1+((1+8)*2)+1)

// This macro allows to define functions with extrachecks to compute the filename
//
#define DECLARE_key2fn(_internal_name,_msg_type,_msg_member)					\
static inline int key2fn_##_internal_name(const msg_iod_t * const msg, char *filename) {	\
	 	int ret;									\
		if ((_msg_type) != msg->msgtype) {						\
			Dprintf(L_WARNING, "invalid request type=0x%04x instead of %x",			\
			msg->msgtype, (_msg_type));						\
			ret = -1;								\
		} else {									\
			(void) snprintf(filename, KEYMAXSIZE, "%2.2x/i%8.8xs%8.8x",		\
					hash(msg->_msg_member.inode, msg->_msg_member.seed),	\
					msg->_msg_member.inode, msg->_msg_member.seed);		\
			ret = 0;								\
		}										\
		return ret;									\
}

// Declare the key2fn_* functions
// For instance, the read function is:
//   static inline int key2fn_read(const msg_iod_t * const msg, char *filename);
DECLARE_key2fn(read,NFSP_READ_REQ,readreq)
DECLARE_key2fn(write,NFSP_WRITE_REQ,writereq)
#ifdef WRITEMIRROR_SUPPORT
DECLARE_key2fn(writemirror,NFSP_WRITEMIRROR_REQ,writemirrorreq)
#endif /* WRITEMIRROR_SUPPORT */
DECLARE_key2fn(fwdread1,NFSP_FWDREAD_1_REQ,fwdreadreq)
DECLARE_key2fn(fwdread2,NFSP_FWDREAD_2_REQ,fwdreadreq)
DECLARE_key2fn(fwdwrite1,NFSP_FWDWRITE_1_REQ,fwdwritereq)
DECLARE_key2fn(fwdwrite2,NFSP_FWDWRITE_2_REQ,fwdwritereq)


static inline int get_realoffset(int bs, int w, int offset, int count) {
	int realoffset;

	Dprintf(D_IOD, "stripe_bs=%d stripe_w=%d o=%d count=%d", bs, w, offset, count);
	
	// the file will contain appended parts of stripes
	// node 0:    [ (0)   (N)   ...]
	// ...
	// node N-1:  [(N-1) (2N-1) ...]
	if (offset / bs != (offset + count - 1) / bs) {
		Dprintf(L_WARNING, "cross stripe operation requested - this is bad");
		realoffset = -1;
	} else {
		realoffset = ((offset / (w * bs)) * bs) + (offset % bs);
	}

	return realoffset;
}


// Fills every field of 'rpc' using data provided by 'm'
static inline void do_rpc_header(struct myrpc_msg * const rpc, const u_long xid) {
	rpc->rm_xid = xid;
	rpc->rm_direction = htonl(REPLY);
	rpc->rp_stat = htonl(MSG_ACCEPTED);
	rpc->oa_flavor = 0;		// must be NULL
	rpc->oa_base = NULL;		// must be NULL
	rpc->ar_stat = htonl(SUCCESS);
}

// Initialize every field of 'udp' including checksum
// Fills the fields of 'ip' used for checksumming the UDP packet,
// zeroes the other ones
static inline void do_udp_header(struct udphdr * const udp,
 				struct iphdr * const ip,
				const char * const eod,
				const msg_iod_t * const m) {

	// Every field of UDP header is set
	udp->source = m->server_port;
	udp->dest = m->client_port;
	udp->len = htons(eod - ((char*) udp));
	udp->check = 0;

	// Pseudo IP-header: every fields of this struct must be set to 0 but
	// this header will be used for checksumming the UDP packet
	(void) memset((char*) ip, '\0', sizeof(struct iphdr));
	ip->saddr = m->server_addr;
	ip->daddr = m->client_addr;
	ip->protocol = IPPROTO_UDP;
	ip->tot_len = udp->len;

	// Calculate the checksum of the UDP packet
	udp->check = in_cksum((unsigned short *)ip, eod - (char*)ip);
}


// To lower the load of the metaserver, data is sent the same way way it is read.
// Since on x86 we have the bad endianness (LSB) we have to convert the values
// to the good one(c)(tm) that is XDR's == MSB
//
// If data is already XDR compliant then this routine should be replaced by a
// simple: memmove(destination, source, sizeof(fattr))
//
#warning "HTONL(FATTR) ENABLED"
static inline void copy_attributes(fattr * const destination, const fattr * const source) {
	unsigned int *src = (unsigned int*) source;
	unsigned int *dst = (unsigned int*) destination;
	unsigned int i = sizeof(fattr) / sizeof(unsigned int);	// number of 4byte words to copy
//	assert(sizeof(fattr) % sizeof(unsigned int) == 0);
	while (i--) {
		*dst = htonl(*src);
		src++;
		dst++;
	}
}


// m		IN	the read message that has to be processed
// Effect : do the read and sends the reply to the client
void do_reply_read(const msg_iod_t * const m) {
	char fn[KEYMAXSIZE];		// will contain the filename

	char b[UDP_MAXLEN];		// will contain all the packet
	struct iphdr *ip;
	struct udphdr *udp;
	struct myrpc_msg *rpc;
	struct readres *nfsres;
	char *endofdata = NULL;		// used to know where the data stops

	int fd;
	int tmp;

	// Get a stripe name
	if (-1 == key2fn_read(m, fn)) {
		return;
	}

	Dprintf(D_IOD, "*do_reply_read()...  about to read from [%s]", fn);

	// Zero-ing the headers
	memset(b, '\0', sizeof(struct iphdr) + sizeof(struct udphdr) +
		       	sizeof(struct myrpc_msg) + sizeof(struct readres));

	// Buffer b will contain :
	//   IP Header / UDP Header / RPC Header / NFS Header
	ip = (struct iphdr *) b;
	udp = (struct udphdr *) ( (char*)ip + sizeof(struct iphdr) );
	rpc = (struct myrpc_msg *) ( (char*)udp + sizeof(struct udphdr));
	nfsres = (struct readres *) ((char*)rpc + sizeof(struct myrpc_msg));

	// 1st part : NFS Header
	// Some sanity checks
	if (!m->readreq.datalen) {
		Dprintf(L_WARNING, "datalen is 0");
	}

#ifdef PARANOID
	// These checks are normally done on the server so this case should not
	// happen
	if ((m->readreq.offset % BLOCKSIZE) + m->readreq.datalen > BLOCKSIZE) {
		Dprintf(L_WARNING, "(offset[%d]%%BLOCKSIZE[%d]) + datalen[%d] = %d > BLOCKSIZE",
			m->readreq.offset, BLOCKSIZE, m->readreq.datalen,
			(m->readreq.offset % BLOCKSIZE) + m->readreq.datalen);
	}
#endif

	fd = open(fn, O_RDONLY);
	if (-1 == fd) {
		if (ENOENT == errno) {
			// We got a request but there is no file
			//  => this is a sparse file => we return an array of 0's
			Dprintf(D_IOD, "sparse file => return %d 0's for %s",
					m->readreq.datalen, fn);

			tmp = m->readreq.datalen;
			memset(&(nfsres->readres_u.reply.data.data_val), '\0', tmp);
			nfsres->status = NFS_OK;
			nfsres->readres_u.reply.data.data_len = htonl(tmp);
			if (tmp & 3) {
				// We have to pad the datas to a 4-bytes multiple
				int pad = 4 - (tmp & 3);
				memset(&(nfsres->readres_u.reply.data.data_val) + tmp, '\0', pad);
				tmp += pad;
			}

			copy_attributes(&(nfsres->readres_u.reply.attributes),
							&(m->readreq.attributes));
			endofdata = (char*)(&(nfsres->readres_u.reply.data.data_val)) + tmp;

		} else {
			// XXX What to do ?
			Dprintf(L_WARNING, "cannot open read only errno=%d", errno);
		}
	} else {
		int realoffset;
		int count = m->readreq.datalen;

		Dprintf(D_IOD, "reading %d bytes from %s", count, fn);

		realoffset = get_realoffset(m->readreq.stripe_bs, m->readreq.stripe_w, m->readreq.offset, count);
		
		if (realoffset != -1) {
			tmp = pread(fd, &(nfsres->readres_u.reply.data.data_val), count,
			    realoffset);
		} else {
			tmp = -1;
		}
		if (-1 == tmp) {
			// The read operation failed
			Dprintf(L_WARNING, "read failed");
			nfsres->status = NFSERR_IO;
			tmp = 0;

		} else {
			// read success
			nfsres->status = NFS_OK;
			nfsres->readres_u.reply.data.data_len = htonl(tmp);
			if (tmp & 3) {
				// We have to pad the datas to a 4-bytes multiple
				int pad = 4 - (tmp & 3);
				memset(&(nfsres->readres_u.reply.data.data_val) + tmp,
						'\0', pad);
				tmp += pad;
			}
			copy_attributes(&(nfsres->readres_u.reply.attributes),
							&(m->readreq.attributes));
		}
		endofdata = (char*)(&(nfsres->readres_u.reply.data.data_val)) + tmp;
		if (-1 == close(fd)) {
			// Weird error...
			Dprintf(L_WARNING, "cannot close() errno=%d", errno);
		}
	}

	// 2nd part : RPC header
	do_rpc_header(rpc, m->readreq.xid);

	// 3rd part : UDP header + pseudo ip header
	do_udp_header(udp, ip, endofdata, m);

	// 4th & last part : send the IP packets on the wire
	switch (xmitmode) {
#if HAS_ROOT
		case XMITMODE_ARP:
			ETHERNET_send(ip, llsocket, m->client_mac_addr);
			break;
		case XMITMODE_IP:
			IP_send(ip, iprawsocket);
			break;
#endif /* HAS_ROOT */
		case XMITMODE_UDP:
			UDP_send(ip, udpsocket);
			break;
		default:
			Dprintf(L_WARNING, "Internal error: unknown xmitmode %d", xmitmode);
	}
}


// -------------------------------------------------------------------------------------------

static void do_forward_step1_read(msg_iod_t * m) {
	char fn[KEYMAXSIZE];		// will contain the filename
	int tmp;
	int ret;

	int readlen;
	struct sockaddr_in nextiod;
	int fd;

	if (-1 == key2fn_fwdread1(m, fn)) {
		return;
	}
	Dprintf(D_IOD, "*do_forward_step1_read()...  about to read from [%s]", fn);

	// Grab and sets values in m for forwarding
	
	if (m->fwdreadreq.offset % BLOCKSIZE + m->fwdreadreq.datalen <= BLOCKSIZE) {
		Dprintf(L_WARNING, "incorrect request in do_forward_step1_read");
		return;
	}
	
	readlen = BLOCKSIZE - (m->fwdreadreq.offset % BLOCKSIZE);
	m->fwdreadreq.nextlen = m->fwdreadreq.datalen - readlen;

	memcpy(&nextiod, &(m->fwdreadreq.nextiod), sizeof(nextiod));
	memset(&(m->fwdreadreq.nextiod), '\0', sizeof(struct sockaddr_in));

	fd = open(fn, O_RDONLY);
	if (-1 == fd) {
		if (ENOENT == errno) {
			// We got a request but there is no file
			//  => this is a sparse file => we return an array of 0's
			Dprintf(D_IOD, "sparse file => return %d 0's for %s", readlen, fn);

			memset(&(m->fwdreadreq.dataval[0]), '\0', readlen);
			tmp = readlen;
		} else {
			Dprintf(L_WARNING, "do_forward_step1_read(): cannot open read only errno=%d", errno);
			tmp = 0;
		}
	} else {
		int realoffset;
		int count = readlen;
		Dprintf(D_IOD, "reading %d bytes from %s", count, fn);

		realoffset = get_realoffset(m->fwdreadreq.stripe_bs, m->fwdreadreq.stripe_w, m->fwdreadreq.offset, count);
		if (realoffset != -1) {
			tmp = pread(fd, &(m->fwdreadreq.dataval[0]), count, realoffset);
		} else {
			tmp = -1;
		}
		

		// read was not complete
		if (tmp != readlen) {
			if (-1 == tmp) {
				// read failure
				Dprintf(L_WARNING, "do_forward_step1_read(): read failed errno=%d", errno);
				tmp = 0;
			} else {
				// read success but EOF
				Dprintf(L_WARNING, "do_forward_step1_read(): read only %d/%d bytes", tmp, readlen);
				memset(&(m->fwdreadreq.dataval[tmp]), '\0', readlen - tmp);
				tmp = readlen;
			}
		}
		if (-1 == close(fd)) {
			// Weird error...
			Dprintf(L_WARNING, "do_forward_step1_read(): cannot close() errno=%d", errno);
		}
	}

	// Changing the type to a inter-iod forwarded read request
	m->msgtype = NFSP_FWDREAD_2_REQ;

	Dprintf(D_IOD, "do_forward_step1_read(): OK");
	ret = sendto(forward_step1_read_udpsock, m, MSG_IOD_FWDREAD_LEN(tmp), 0,
			&nextiod, sizeof(struct sockaddr_in));
	if (ret == -1) {
		Dprintf(L_WARNING, "do_forward_step1_read(): cannot send UDP packet errno=%d", errno);
	}
}



// m		IN	the read message that has to be processed
// Effect : do the last part of the read and sends the reply back to the client
void do_forward_step2_read(msg_iod_t * const m) {
	char fn[KEYMAXSIZE];		// will contain the filename

	char b[UDP_MAXLEN];		// will contain all the packet
	struct iphdr *ip;
	struct udphdr *udp;
	struct myrpc_msg *rpc;
	struct readres *nfsres;
	char *readbuf = NULL;
	char *endofdata = NULL;		// used to know where the data stops

	int fd;
	int tmp;


	if (-1 == key2fn_fwdread2(m, fn)) {
		return;
	}

	Dprintf(D_IOD, "*do_forward_step2_read()...  about to read from [%s]", fn);

#if 1|| defined(PARANOID)
	// Check that nextiod field has been Zero'ed
	do {
		struct sockaddr_in blank;
		memset(&blank, '\0', sizeof(struct sockaddr_in));
		if (memcmp(&(m->fwdreadreq.nextiod), &blank, sizeof(struct sockaddr_in))) {
			Dprintf(L_WARNING, "do_forward_step2_read: nextiod should be zero => ignoring req");
			return;
		}

		// If we're here then we should have a request spanning on several BLOCK
		if ((m->fwdreadreq.offset % BLOCKSIZE) + m->fwdreadreq.datalen <= BLOCKSIZE) {
			Dprintf(L_WARNING, "do_forward_step2_read: request does not span on several blocks => ignoring req");
			Dprintf(L_WARNING, "(offset[%d]%%BLOCKSIZE[%d]) + datalen[%d] = %d <= BLOCKSIZE",
				m->fwdreadreq.offset, BLOCKSIZE, m->fwdreadreq.datalen,
				(m->fwdreadreq.offset % BLOCKSIZE) + m->fwdreadreq.datalen);
			return;
		}

		// It's not critical but weird
		if (0 == m->fwdreadreq.datalen) {
			Dprintf(L_WARNING, "WARNING: datalen is 0");
		}

	} while(0);
#endif

	// Buffer b will contain :
	//   IP Header / UDP Header / RPC Header / NFS Header
	ip = (struct iphdr *) b;
	udp = (struct udphdr *) ( (char*)ip + sizeof(struct iphdr) );
	rpc = (struct myrpc_msg *) ( (char*)udp + sizeof(struct udphdr));
	nfsres = (struct readres *) ((char*)rpc + sizeof(struct myrpc_msg));

	// Zero-ing the headers of the reply message
	memset(b, '\0', sizeof(struct iphdr) + sizeof(struct udphdr) +
		       	sizeof(struct myrpc_msg) + sizeof(struct readres));

	Dprintf(D_IOD, "Shifting by %d bytes", m->fwdreadreq.datalen - m->fwdreadreq.nextlen);
	// A payload of m->fwdreadreq.datalen - m->fwdreadreq.nextlen should be found
	// so the offset of the buffer is shifted by this value.
	memcpy(&(nfsres->readres_u.reply.data.data_val),
		   &(m->fwdreadreq.dataval),
		   m->fwdreadreq.datalen - m->fwdreadreq.nextlen);

	// readbuf is the buf in the message where we can complete the buffer
	readbuf = ((char *) &(nfsres->readres_u.reply.data.data_val))
		+ (m->fwdreadreq.datalen - m->fwdreadreq.nextlen);

	// 1st part : NFS Header
	// Note: at the end, tmp corresponds to the length of data to send back to the
	// client
	fd = open(fn, O_RDONLY);
	if (-1 == fd) {
		if (ENOENT == errno) {
			// Request but no file => sparse file => we return an array of 0's
			Dprintf(D_IOD, "sparse file => return %d 0's for %s", m->fwdreadreq.datalen, fn);

			tmp = m->fwdreadreq.datalen;
			memset(readbuf, '\0', m->fwdreadreq.nextlen);
			nfsres->status = NFS_OK;
			nfsres->readres_u.reply.data.data_len = htonl(tmp);
			if (tmp & 3) {
				// We have to pad the datas to a 4-bytes multiple
				int pad = 4 - (tmp & 3);
				memset(&(nfsres->readres_u.reply.data.data_val) + tmp, '\0', pad);
				tmp += pad;
			}

			copy_attributes(&(nfsres->readres_u.reply.attributes),
							&(m->fwdreadreq.attributes));

		} else {
			Dprintf(L_WARNING, "cannot open read only errno=%d", errno);
			nfsres->status = NFSERR_IO;
			tmp = 0;
		}
	} else {
		int realoffset;
		int count = m->fwdreadreq.nextlen;
		Dprintf(D_IOD, "reading count=nextlen=%d bytes from %s", count, fn);

		realoffset = get_realoffset(m->fwdreadreq.stripe_bs, m->fwdreadreq.stripe_w,
				m->fwdreadreq.offset + (m->fwdreadreq.datalen - m->fwdreadreq.nextlen),
				count);

		if (realoffset != -1) {
			tmp = pread(fd, readbuf, count, realoffset);
		} else {
			tmp = -1;
		}

		if (-1 == tmp) {
			// The read operation failed
			Dprintf(L_WARNING, "read failed");
			nfsres->status = NFSERR_IO;
			tmp = 0;
		} else {
			tmp += m->fwdreadreq.datalen - m->fwdreadreq.nextlen;
			// read success
			nfsres->status = NFS_OK;
			nfsres->readres_u.reply.data.data_len = htonl(tmp);
			if (tmp & 3) {
				// We have to pad the datas to a 4-bytes multiple
				int pad = 4 - (tmp & 3);
				memset(&(nfsres->readres_u.reply.data.data_val)+tmp, '\0', pad);
				tmp += pad;
			}
			copy_attributes(&(nfsres->readres_u.reply.attributes),
							&(m->readreq.attributes));
		}
		if (-1 == close(fd)) {
			Dprintf(L_WARNING, "cannot close() errno=%d", errno);
		}
	}

	Dprintf(D_IOD, "Full DATA size is %dB", tmp);
	endofdata = ((char *) &(nfsres->readres_u.reply.data.data_val)) + tmp;

	// 2nd part : RPC header
	do_rpc_header(rpc, m->readreq.xid);

	// 3rd part : UDP header + pseudo ip header
	do_udp_header(udp, ip, endofdata, m);

	// 4th & last part : send the IP packets on the wire
	switch (xmitmode) {
#if HAS_ROOT
		case XMITMODE_ARP:
			ETHERNET_send(ip, llsocket, m->client_mac_addr);
			break;
		case XMITMODE_IP:
			IP_send(ip, iprawsocket);
			break;
#endif /* HAS_ROOT */
		case XMITMODE_UDP:
			UDP_send(ip, udpsocket);
			break;
		default:
			Dprintf(L_WARNING, "Internal error: unknown xmitmode %d", xmitmode);
	}

}


static void do_forward_step1_write(msg_iod_t * m) {
	char fn[KEYMAXSIZE];		// will contain the filename
	int tmp;
	int ret;

	int writelen;
	struct sockaddr_in nextiod;
	int fd;

	if (-1 == key2fn_fwdwrite1(m, fn)) {
		return;
	}
	Dprintf(D_IOD, "*do_forward_step1_write()...  about to write to [%s]", fn);

	// Grab and sets values in m for forwarding
	
	if (m->fwdwritereq.offset % BLOCKSIZE + m->fwdwritereq.datalen <= BLOCKSIZE) {
		Dprintf(L_WARNING, "incorrect request in do_forward_step1_write");
		return;
	}
	
	writelen = BLOCKSIZE - (m->fwdwritereq.offset % BLOCKSIZE);
	m->fwdwritereq.nextlen = m->fwdwritereq.datalen - writelen;

	memcpy(&nextiod, &(m->fwdwritereq.nextiod), sizeof(nextiod));

	fd = open(fn, O_WRONLY|O_CREAT, 00600);
	if (-1 == fd) {
		Dprintf(L_WARNING, "cannot open %s WR|CREAT errno=%d", fn, errno);
		// XXX What to do ?
		tmp = -1;
	} else {
		int realoffset;
		int count = writelen;

		realoffset = get_realoffset(m->fwdwritereq.stripe_bs,
				m->fwdwritereq.stripe_w, m->fwdwritereq.offset,
				count);

		if (realoffset != -1) {
			tmp = pwrite(fd, &(m->fwdwritereq.dataval[0]), count, realoffset);
		} else {
			tmp = -1;
		}

		if (tmp != writelen) {	
			if (-1 == tmp) {
				Dprintf(L_WARNING, "write failed");
				// XXX what should be done ?
			} else {
				Dprintf(L_WARNING, "do_forward_step1_write(): short write ???");
				// XXX ???
			}
		}

		if (-1 == close(fd)) {
			// Weird error...
			Dprintf(L_WARNING, "do_forward_step1_write(): cannot close() errno=%d", errno);
		}
	}

	// Changing the type to a inter-iod forwarded write request
	m->msgtype = NFSP_FWDWRITE_2_REQ;

	memmove(&(m->fwdwritereq.dataval[0]),  &(m->fwdwritereq.dataval[writelen]), m->fwdwritereq.nextlen);

	Dprintf(D_IOD, "do_forward_step1_write(): first part written");

	
	ret = sendto(forward_step1_write_udpsock, m, MSG_IOD_FWDWRITE_LEN(m->fwdwritereq.nextlen),
			0,
			(const struct sockaddr *)&nextiod, sizeof(nextiod));
	Dprintf(L_WARNING, NTODSTR, NTOD(nextiod));
	if (ret == -1) {
		Dprintf(L_WARNING, "do_forward_step1_write(): cannot send UDP packet sock=%d m=%p len=%d errno=%d", forward_step1_write_udpsock, m, MSG_IOD_FWDWRITE_LEN(m->fwdwritereq.nextlen), errno);
	}
}


void do_forward_step2_write(msg_iod_t * const m) {
	char fn[KEYMAXSIZE];		// will contain the filename

	char b[UDP_MAXLEN];		// will contain all the packet
	struct iphdr *ip;
	struct udphdr *udp;
	struct myrpc_msg *rpc;
	struct attrstat *nfsres;
	char *endofdata = NULL;		// used to know where the data stops

	int fd;
	int tmp;

	if (-1 == key2fn_fwdwrite2(m, fn)) {
		Dprintf(L_WARNING, "incorrect request in do_forward_step1_write");
		return;
	}

	Dprintf(D_IOD, "*do_forward_step2_write()...  about to write the %d bytes left to [%s]",
			m->fwdwritereq.nextlen, fn);

#if 1|| defined(PARANOID)
	// Check that nextiod field has been Zero'ed
	// If we're here then we should have a request spanning on several BLOCK
	if ((m->fwdwritereq.offset % BLOCKSIZE) + m->fwdwritereq.datalen <= BLOCKSIZE) {
		Dprintf(L_WARNING, "do_forward_step2_write: request does not span on several blocks => ignoring req");
		Dprintf(L_WARNING, "(offset[%d]%%BLOCKSIZE[%d]) + datalen[%d] = %d <= BLOCKSIZE",
			m->fwdwritereq.offset, BLOCKSIZE, m->fwdwritereq.datalen,
			(m->fwdwritereq.offset % BLOCKSIZE) + m->fwdwritereq.datalen);
		return;
	}

	// It's not critical but weird
	if (0 == m->fwdwritereq.datalen) {
		Dprintf(L_WARNING, "WARNING: datalen is 0");
	}

#endif

	// Buffer b will contain :
	//   IP Header / UDP Header / RPC Header / NFS Header
	ip = (struct iphdr *) b;
	udp = (struct udphdr *) ( (char*)ip + sizeof(struct iphdr) );
	rpc = (struct myrpc_msg *) ( (char*)udp + sizeof(struct udphdr));
	nfsres = (struct attrstat *) ((char*)rpc + sizeof(struct myrpc_msg));

	// Zero-ing the headers of the reply message
	memset(b, '\0', sizeof(struct iphdr) + sizeof(struct udphdr) +
		       	sizeof(struct myrpc_msg) + sizeof(struct readres));

	// 1st part : NFS Header
	// Note: at the end, tmp corresponds to the length of data to send back to the
	// client
	fd = open(fn, O_WRONLY|O_CREAT, 00600);
	if (-1 == fd) {
		Dprintf(L_WARNING, "cannot open O_WRONLY|O_CREAT [%s] errno=%d", fn, errno);
		nfsres->status = NFSERR_IO;
		tmp = 0;
	} else {
		int realoffset;
		int count = m->fwdwritereq.nextlen;

		Dprintf(D_IOD, "writing count=nextlen=%d bytes to [%s]", count, fn);

		realoffset = get_realoffset(m->fwdwritereq.stripe_bs, m->fwdwritereq.stripe_w,
				m->fwdwritereq.offset + m->fwdwritereq.datalen - m->fwdwritereq.nextlen,
				count);

		if (realoffset % BLOCKSIZE) {
			Dprintf(L_WARNING, "INTERNAL ERROR: realoffset is not a multiple of BLOCKSIZE");
		}

		if (realoffset != -1) {
			tmp = pwrite(fd, m->writereq.dataval, count, realoffset);
		} else {
			tmp = -1;
		}

		if (tmp != count) {
			if (-1 == tmp) {
				// The write operation failed
				Dprintf(L_WARNING, "write failed");
				nfsres->status = NFSERR_IO;
				tmp = 0;
			} else {
				Dprintf(L_WARNING, "short write");
				nfsres->status = NFSERR_IO;
				tmp = 0;
			}
		} else {
			nfsres->status = NFS_OK;
			copy_attributes(&(nfsres->attrstat_u.attributes), &(m->fwdwritereq.attributes));
		}
	}

	endofdata = (char*)&(nfsres->attrstat_u.attributes) + sizeof(fattr);
	if (-1 == close(fd)) {
		Dprintf(L_WARNING, "cannot close() errno=%d", errno);
	}

	// 2nd part : RPC header
	do_rpc_header(rpc, m->writereq.xid);

	// 3rd part : UDP header + pseudo ip header
	do_udp_header(udp, ip, endofdata, m);


	// 4th & last part : send the IP packets on the wire
	switch (xmitmode) {
#if HAS_ROOT
		case XMITMODE_ARP:
			ETHERNET_send(ip, llsocket, m->client_mac_addr);
			break;
		case XMITMODE_IP:
			IP_send(ip, iprawsocket);
			break;
#endif /* HAS_ROOT */
		case XMITMODE_UDP:
			UDP_send(ip, udpsocket);
			break;
		default:
			Dprintf(L_WARNING, "Internal error: unknown xmitmode %d", xmitmode);
	}
}


// -------------------------------------------------------------------------------------------
#ifdef WRITEMIRROR_SUPPORT
// The algo is pretty simple:
// exec(msg->req)
// msg->idx--
// if (msg->idx < 0) {
// 	that's done
// 	} else {
// send(msg->nextiods[idx], msg) //   }
//

void do_reply_writemirror(msg_iod_t * const m) {
	char fn[KEYMAXSIZE];		// will contain the filename

	char b[UDP_MAXLEN];		// will contain the full NFS reply packet
	struct iphdr *ip;
	struct udphdr *udp;
	struct myrpc_msg *rpc;
	struct attrstat *nfsres;
	char *endofdata = NULL;		// used to know where the data stops

	int fd;
	int tmp;

	if (-1 == key2fn_writemirror(m, fn)) {
		return;
	}

	Dprintf(D_IOD, "*do_reply_writemirror()...  about to writemirror to [%s]", fn);

	// Sanity check: sensible index
	if (m->writemirrorreq.idx < -1 || m->writemirrorreq.idx > MAX_NEXTIODS - 1) {
		Dprintf(L_WARNING, "invalid request (idx <0 || idx > MAX_NEXTIODS-1)");
		return;
	}
	Dprintf(D_IOD, "*do_reply_writemirror()...  idx=%d", m->writemirrorreq.idx);

#if 0
	do {
		int i;

		for (i=0; i<MAX_NEXTIODS; i++) {
#warning "inet_aton is not MT-SAFE"			
			Dprintf(D_IOD, "*do_reply_writemirror()... nextiods[%d] == %s:%d\n", i,
					inet_ntoa(m->writemirrorreq.nextiods[i].sin_addr),
					ntohs(m->writemirrorreq.nextiods[i].sin_port),
					MSG_IOD_PING_LEN);

		}
	} while (0);
#endif 
	 

	// Sanity check: a non-zero operation is wiser
	if (m->writemirrorreq.datalen == 0) {
		Dprintf(L_WARNING, "weird packet: datalen is 0 => request dropped");
		return;
	}

	// Sanity check: we do not handle writes when they're located on block boundaries
	if ((m->writemirrorreq.offset % BLOCKSIZE) + m->writemirrorreq.datalen 
		> BLOCKSIZE) {
		Dprintf(L_WARNING, "(offset[%d]%%BLOCKSIZE[%d]) + datalen[%d] = %d > BLOCKSIZE",
			m->writemirrorreq.offset, BLOCKSIZE,
			m->writemirrorreq.datalen,
			(m->writemirrorreq.offset % BLOCKSIZE) + m->writemirrorreq.datalen);
		return;
	}



	// Zero-ing the headers
	memset(b, '\0', sizeof(b));

	// Buffer b will contain :
	//   IP Header / UDP Header / RPC Header / NFS Header
	ip = (struct iphdr *) b;
	udp = (struct udphdr *) ( (char*)ip + sizeof(struct iphdr) );
	rpc = (struct myrpc_msg *) ( (char*)udp + sizeof(struct udphdr));
	nfsres = (struct attrstat *) ((char*)rpc + sizeof(struct myrpc_msg));

	// 1st part : NFS Header
	Dprintf(D_IOD, "Writing %d bytes to %s", m->writemirrorreq.datalen, fn);
	fd = open(fn, O_WRONLY|O_CREAT, 00600);
	if (-1 == fd) {
		Dprintf(L_WARNING, "cannot open %s WR|CREAT errno=%d", fn, errno);
		// XXX What to do ?
		tmp = -1;
	} else {
		int realoffset;
		int count = m->writemirrorreq.datalen;

		realoffset = get_realoffset(m->writemirrorreq.stripe_bs,
				m->writemirrorreq.stripe_w,
				m->writemirrorreq.offset, count);

		if (realoffset != -1) {
			tmp = pwrite(fd, m->writemirrorreq.dataval, count, realoffset);
		} else {
			tmp = -1;
		}
	}
	
	if (-1 == tmp) {
		// The write operation failed
		Dprintf(L_WARNING, "something failed during writemirror");
		nfsres->status = NFSERR_IO;
		// XXX what should be done ?
	} else {
		nfsres->status = NFS_OK;
		copy_attributes(&(nfsres->attrstat_u.attributes), &(m->writemirrorreq.attributes));
	}

	endofdata = (char*)&(nfsres->attrstat_u.attributes) + sizeof(fattr);

	if (-1 == close(fd)) {
		// Weird error... It might happen according to the man page but what to do in such
		// case ? Currently we can only do a warning as anything could have happened...
		Dprintf(L_WARNING, "cannot close() errno=%d", errno);
	}


	// We'll now decide what to do with this peculiar request
	// (Note: this comment is duplicated in iodng.h)
	// 
	// If idx == -1 then the iod must return the NFS ack to the client
	// If idx >=0 then the iod forwards the data to the
	//   next iod in the list that is nextiods[idx] (and
	//   decrease the idx in the message)
	//
	// Example:
	//   S: Serveur
	//   I1, I2, I3: iods
	//   In the message, nextiods[0]=I2 nextiods[1]=I3
	//
	//   S to I1:  idx=1   send to nextiods[idx] with idx set to 1-1
	//   I1 to I3: idx=0   send to nextiods[idx] with idx set to 0-1
	//   I3 to I2: idx=-1  send NFS ack to client
	if (m->writemirrorreq.idx >= 0) {
		struct sockaddr_in sin;
		int sin_len = sizeof(sin);

		// there are still otherthe message has to be forwarded so we'll reuse the mirror
		// idx >= 1
		// NB: we use the source message to avoid a memory copy

		memcpy(&sin, &(m->writemirrorreq.nextiods[m->writemirrorreq.idx]), sizeof(sin));

		m->writemirrorreq.idx--;

		if (-1 == sendto(write_mirror_udpsock, m,
					MSG_IOD_WRITE_MIRROR_LEN(m->writemirrorreq.datalen),
					MSG_NOSIGNAL, &sin, sin_len)) {
			Dprintf(L_WARNING, "sendto failed on a UDP socket errno=%d", errno);
		}
		return;
	}

	if (nfsres->status != NFS_OK) {
		Dprintf(D_IOD, "a NFS error occured => stopping the forwarding and replying a NACK to the client");
	}
	if (m->writemirrorreq.idx == 0) {
		Dprintf(D_IOD, "final index so generating NFS reply");
	}

	// 2nd part : RPC header
	do_rpc_header(rpc, m->writemirrorreq.xid);

	// 3rd part : UDP header + pseudo ip header
	do_udp_header(udp, ip, endofdata, m);

	// 4th & last part : send the IP packets on the wire
	switch (xmitmode) {
#if HAS_ROOT
		case XMITMODE_ARP:
			ETHERNET_send(ip, llsocket, m->client_mac_addr);
			break;
		case XMITMODE_IP:
			IP_send(ip, iprawsocket);
			break;
#endif /* HAS_ROOT */
		case XMITMODE_UDP:
			UDP_send(ip, udpsocket);
			break;
		default:
			Dprintf(L_WARNING, "Internal error: unknown xmitmode %d", xmitmode);
	}

}
#endif /* WRITEMIRROR_SUPPORT */
// -------------------------------------------------------------------------------------------


void do_reply_write(const msg_iod_t * const m) {
	char fn[KEYMAXSIZE];		// will contain the filename

	char b[UDP_MAXLEN];		// will contain all the packet
	struct iphdr *ip;
	struct udphdr *udp;
	struct myrpc_msg *rpc;
	struct attrstat *nfsres;
	char *endofdata = NULL;		// used to know where the data stops

	int fd;
	int tmp;

	if (-1 == key2fn_write(m, fn)) {
		return;
	}
	Dprintf(D_IOD, "*do_reply_write()...  about to write to [%s]", fn);

	// Zero-ing the headers
	memset(b, '\0', sizeof(b));

	// Buffer b will contain :
	//   IP Header / UDP Header / RPC Header / NFS Header
	ip = (struct iphdr *) b;
	udp = (struct udphdr *) ( (char*)ip + sizeof(struct iphdr) );
	rpc = (struct myrpc_msg *) ( (char*)udp + sizeof(struct udphdr));
	nfsres = (struct attrstat *) ((char*)rpc + sizeof(struct myrpc_msg));

	// 1st part : NFS Header
	// Some sanity checks
	if (!m->writereq.datalen) {
		Dprintf(L_WARNING, "datalen is 0");
	}

	if ((m->writereq.offset % BLOCKSIZE) + m->writereq.datalen > BLOCKSIZE) {
		Dprintf(L_WARNING, "(offset[%d]%%BLOCKSIZE[%d]) + datalen[%d] = %d > BLOCKSIZE",
				m->writereq.offset,	BLOCKSIZE,
				m->writereq.datalen,
				(m->writereq.offset % BLOCKSIZE) + m->writereq.datalen);
	}

	Dprintf(D_IOD, "Writing %d bytes to %s", m->writereq.datalen, fn);
	fd = open(fn, O_WRONLY|O_CREAT, 00600);
	if (-1 == fd) {
		Dprintf(L_WARNING, "cannot open %s WR|CREAT errno=%d", fn, errno);
		// XXX What to do ?
		tmp = -1;
	} else {
		int realoffset;
		int count = m->writereq.datalen;

		realoffset = get_realoffset(m->writereq.stripe_bs,
				m->writereq.stripe_w, m->writereq.offset,
				count);

		if (realoffset != -1) {
			tmp = pwrite(fd, m->writereq.dataval, count, realoffset);
		} else {
			tmp = -1;
		}
	}
	
	if (-1 == tmp) {
		// The write operation failed
		Dprintf(L_WARNING, "write failed");
		nfsres->status = NFSERR_IO;
		// XXX what should be done ?
	} else {
		nfsres->status = NFS_OK;
		copy_attributes(&(nfsres->attrstat_u.attributes), &(m->writereq.attributes));
	}

	endofdata = (char*)&(nfsres->attrstat_u.attributes) + sizeof(fattr);

	if (-1 == close(fd)) {
		// Weird error...
		Dprintf(L_WARNING, "cannot close() errno=%d", errno);
	}

	// 2nd part : RPC header
	do_rpc_header(rpc, m->writereq.xid);

	// 3rd part : UDP header + pseudo ip header
	do_udp_header(udp, ip, endofdata, m);

	// 4th & last part : send the IP packets on the wire
	switch (xmitmode) {
#if HAS_ROOT
		case XMITMODE_ARP:
			ETHERNET_send(ip, llsocket, m->client_mac_addr);
			break;
		case XMITMODE_IP:
			IP_send(ip, iprawsocket);
			break;
#endif /* HAS_ROOT */
		case XMITMODE_UDP:
			UDP_send(ip, udpsocket);
			break;
		default:
			Dprintf(L_WARNING, "Internal error: unknown xmitmode %d", xmitmode);
	}
}


// This functions removes the files described in m
void do_purge(const msg_iod_t * const m) {
	int idx;		// index of data file to remove in 'm'
	char file[KEYMAXSIZE];

	Dprintf(D_IOD, "*do_purge()...");

	for (idx = 0; idx < m->purgereq.nfiles; idx++) {
		unsigned int inode, seed;
		
		inode = m->purgereq.files[idx].inode;
		if (inode == 0) {
			// 0 is an invalid inode (and is used as padding)
			continue;
		}
		seed = m->purgereq.files[idx].seed;

		sprintf(file, "%2.2x/i%8.8xs%8.8x", hash(inode, seed), inode, seed);
		
		Dprintf(D_IOD, "*do_purge()...  about to unlink [%s]", file);

		if (-1 == unlink(file)) {
			Dprintf(L_WARNING, "error while unlinking '%s' errno=%d", file, errno);
		}
	}

}


void do_reply_ping_spoof(const msg_iod_t * const msrc) {
	char b[UDP_MAXLEN];		// will contain all the packet
	struct iphdr *ip;
	struct udphdr *udp;
	msg_iod_t *msg;
	char *endofdata = NULL;		// used to know where the data stops

	Dprintf(D_IOD, "*do_reply_ping_spoof()...");

	// Buffer b will contain :
	//   IP Header / UDP Header / msg_iod
	ip = (struct iphdr *) b;
	udp = (struct udphdr *) ((char*)ip + sizeof(struct iphdr));
	msg = (msg_iod_t *) ((char*)udp + sizeof(struct udphdr));

	// Setup of the PONG reply
	msg->magic = NFSP_MAGIC_MSG;
	msg->msgtype = NFSP_PONG_REQ;
	msg->client_addr = msrc->client_addr;
	msg->server_addr = msrc->server_addr;
	
	// Copy the MAC address
	memcpy(msg->client_mac_addr, msrc->client_mac_addr, 6);
	
	msg->client_port = msrc->client_port;
	msg->server_port = msrc->server_port;
	msg->pingreq.seq = msrc->pingreq.seq;

	endofdata = (char*) msg + MSG_IOD_PING_LEN;

	do_udp_header(udp, ip, endofdata, msg);

	// send the IP packets on the wire
	switch (xmitmode) {
#if HAS_ROOT
		case XMITMODE_ARP:
			ETHERNET_send(ip, llsocket, msg->client_mac_addr);
			break;
		case XMITMODE_IP:
			IP_send(ip, iprawsocket);
			break;
#endif /* HAS_ROOT */
		case XMITMODE_UDP:
			UDP_send(ip, udpsocket);
			break;
		default:
			Dprintf(L_WARNING, "Internal error: unknown xmitmode %d", xmitmode);
	}
}


// Handles READ/WRITE/PING requests
// Ideally priority should be the lowest
void* IO_thr(void *dummy) {
	int i;

	Dprintf(D_IOD, "IO_thr(pid=%d): started as uid=%d euid=%d gid=%d egid=%d",
	     getpid(), getuid(), geteuid(), getgid(), getegid());

#if HAS_PRIO
	if (-1 == seteuid(0)) {
		Dprintf(L_FATAL, "cannot change to uid 0");
		exit(1);
	}

	if (-1 == setpriority(PRIO_PROCESS, getpid(), PRIO_IO_THR)) {
		Dprintf(L_FATAL, "setpriority(IO_thr,%d) failed errno=%d", PRIO_IO_THR, errno);
		exit(1);
	}
	if (-1 == setregid(user->pw_gid, user->pw_gid)) {
		Dprintf(L_FATAL, "cannot change to %s gid", user->pw_name);
		exit(1);
	}
	if (-1 == setreuid(user->pw_uid, user->pw_uid)) {
		Dprintf(L_FATAL, "cannot change to %s uid", user->pw_name);
		exit(1);
	}
	
	Dprintf(D_IOD, "IO_thr(pid=%d): running as uid=%d euid=%d gid=%d egid=%d",
	     getpid(), getuid(), geteuid(), getgid(), getegid());
#endif /* HAS_PRIO */
	

	while (1) {
		pthread_mutex_lock(&msg_mutex);

		while (msg_filled <= 0) {
			pthread_cond_wait(&msg_cond_filled, &msg_mutex);
		}

		// If we're here then at least 1 slot is marked FILLED

		// Retrieve the first slot filled
		i = iIndexFILLED;

		// The following loop cannot be infinite as there's a FILLED slot
		while (msg_state[i] != FILLED) {
			i = (i + 1) % MSG_ARRAY_SIZE;
		}

		iIndexFILLED = (i + 1) % MSG_ARRAY_SIZE ;

		// We have been signalled there is at least one message in state FILLED
		assert(FILLED == msg_state[i]);

		msg_state[i] = USED;
		msg_filled--;
		pthread_mutex_unlock(&msg_mutex);

		switch (msg_array[i].msgtype) {
			case NFSP_READ_REQ:
				do_reply_read(msg_array + i);
				break;

			case NFSP_WRITE_REQ:
				do_reply_write(msg_array + i);
				break;

#ifdef WRITEMIRROR_SUPPORT
			case NFSP_WRITEMIRROR_REQ:
				do_reply_writemirror(msg_array + i);
				break;
#endif /* WRITEMIRROR_SUPPORT */

			case NFSP_FWDREAD_1_REQ:
				do_forward_step1_read(msg_array + i);
				break;

			case NFSP_FWDREAD_2_REQ:
				do_forward_step2_read(msg_array + i);
				break;

			case NFSP_FWDWRITE_1_REQ:
				do_forward_step1_write(msg_array + i);
				break;

			case NFSP_FWDWRITE_2_REQ:
				do_forward_step2_write(msg_array + i);
				break;

			case NFSP_PURGE_REQ:
				do_purge(msg_array + i);
				break;

			case NFSP_PING_REQ:
				do_reply_ping_spoof(msg_array + i);
				break;

			case NFSP_PONG_REQ:
				Dprintf(L_WARNING, "protocol error: received PONG_REP");
				break;

			default:
				Dprintf(L_WARNING, "unknown message type msg_array[i].msgtype=%d",
						msg_array[i].msgtype);
		}

		pthread_mutex_lock(&msg_mutex);
		msg_state[i] = FREE;
		msg_free++;

		assert(0 < msg_free);
		assert(msg_free <= MSG_ARRAY_SIZE);
		pthread_cond_signal(&msg_cond_free);
		pthread_mutex_unlock(&msg_mutex);
	}
}



// Handles the UDP socket: Grabs UDP packets and store them in msg_array[]
// The read is synchronous and will happen only if some free slots in msg_array
// have been found
void* UDP_thr(void *dummy) {
	int i;
	struct sockaddr_in sockstruct;			// used as a scratch reg.
	socklen_t socklen = sizeof(struct sockaddr_in);	// only read


	Dprintf(D_IOD, "UDP_thr(pid=%d): started as uid=%d euid=%d gid=%d egid=%d",
	     getpid(), getuid(), geteuid(), getgid(), getegid());

#if HAS_PRIO
	if (-1 == setpriority(PRIO_PROCESS, getpid(), PRIO_UDP_THR)) {
		Dprintf(L_FATAL, "setpriority(UDP_thr,%d) failed errno=%d", PRIO_UDP_THR, errno);
		exit(1);
	}
	if (-1 == setregid(user->pw_gid, user->pw_gid)) {
		Dprintf(L_FATAL, "cannot change to %s uid/gid", user->pw_name);
		exit(1);
	}

	Dprintf(D_IOD, "UDP_thr(pid=%d): running as uid=%d euid=%d gid=%d egid=%d",
	     getpid(), getuid(), geteuid(), getgid(), getegid());
#endif /* HAS_PRIO */

	while (1) {
		// We wait for a slot being freed, event which will eventually
		// happen since there are IO threads that free slots
		pthread_mutex_lock(&msg_mutex);
		while (msg_free == 0) {
			pthread_cond_wait(&msg_cond_free, &msg_mutex);
		}

		// If we're here then at least a slot is marked FREE

		// bounded FIFO
		i = iIndexFREE;

		// The following loop cannot be infinite as there's a FREE slot
		while (msg_state[i] != FREE) {
			i = (i + 1) % MSG_ARRAY_SIZE;
		}

		iIndexFREE = (iIndexFREE + 1) % MSG_ARRAY_SIZE;

		// We were signalled there was at least one free slot
		assert(FREE == msg_state[i]);

		// Found a free block so we'll use it
		msg_state[i] = RECV;
		msg_free--;

		// We no longer need the mutex
		pthread_mutex_unlock(&msg_mutex);

		Dprintf(D_IOD, "slot %d: FREE=>RECV [UDP_thr]", i);

		/************************************************************************/
		do {
			int bytesread;
retryread:
			Dprintf(D_IOD, "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -");
			// Since we are the only thread allowed to alter the state of slot
			// i from RECV to FILLED we do not need locking here
			bytesread = recvfrom(udpsocket, msg_array + i,
					MSG_IOD_LEN_MAX, 0,
					(struct sockaddr *)&sockstruct,
					&socklen);

			Dprintf(D_IOD, "Received %d bytes from " NTODSTR " (UDP) in slot=%d/%d",
				bytesread, NTOD(sockstruct), i, MSG_ARRAY_SIZE);
				
			// Do some simplistic checks ...
			// ... about the message size
			if (bytesread < (int) MSG_IOD_COMMON_LEN) {
				if (bytesread < 0) {
					Dprintf(L_WARNING, "recvfrom() errno=%d", errno);
				} else {
					Dprintf(L_WARNING, "invalid request (too small message)");
				}
				goto retryread;
			} else if (bytesread > (int) MSG_IOD_LEN_MAX) {
				Dprintf(L_WARNING, "invalid request (too big !)");
				goto retryread;
			}

			// ... about the magic number the message should have
			if (msg_array[i].magic != NFSP_MAGIC_MSG) {
				Dprintf(L_WARNING, "invalid request (not enough magic: 0x%04x instead of 0x%04x)",
						msg_array[i].magic, NFSP_MAGIC_MSG);
				goto retryread;
			}

#if 0 && defined(PARANOID)
			if ((WRITE_REQ == msg_array[i].kind) && \
				(0 == msg_array[i].server_port ||
				 0 == msg_array[i].client_addr ||
				 0 == msg_array[i].client_port)) {
				Dprintf(L_WARNING, "READ_REQ: bad request server_port=%d client_addr=%x client_port=%d",
					ntohs(msg_array[i].server_port),
					ntohl(msg_array[i].client_addr),
					ntohs(msg_array[i].client_port));
				goto retryread;
			}
			if ((READ_REQ == msg_array[i].kind) && \
				(0 == msg_array[i].server_port ||
				 0 == msg_array[i].client_addr ||
				 0 == msg_array[i].client_port)) {
				Dprintf(L_WARNING, "READ_REQ: bad request server_port=%d client_addr=%x client_port=%d",
				ntohs(msg_array[i].server_port),
				ntohl(msg_array[i].client_addr),
				ntohs(msg_array[i].client_port));
				goto retryread;
			}
#endif

			if (0 == msg_array[i].client_addr) {
				Dprintf(D_IOD, "filling client_addr");
				msg_array[i].client_addr = sockstruct.sin_addr.s_addr;
			}
			if (0 == msg_array[i].client_port) {
				Dprintf(D_IOD, "filling client_port");
				msg_array[i].client_port = sockstruct.sin_port;
			}
			if (0 == msg_array[i].server_addr) {
				Dprintf(D_IOD, "filling server_addr");
				msg_array[i].server_addr = sockstruct.sin_addr.s_addr;
			}
			if (0 == msg_array[i].server_port) {
				Dprintf(D_IOD, "filling server_port");
				msg_array[i].server_port = htons(31337); //h4ck z3 w0r1d 8)
			}
		} while(0);

		// We need to inform the other threads that some work is waiting

		pthread_mutex_lock(&msg_mutex);
		msg_state[i] = FILLED;
		msg_filled++;

#ifdef FIFO_PROFILING
		do {
			static int threshold = 10;
			if (msg_filled >= iTestBufferSize) {
				Dprintf(L_WARNING, "NOTICE: number of message reached %d - threshold now doubled\n", iTestBufferSize);
				iTestBufferSize *= 2;
			}
		} while(0);
#endif

		assert(0 < msg_filled);
		assert(msg_filled <= MSG_ARRAY_SIZE);

		// Wake up an I/O thread
		pthread_cond_signal(&msg_cond_filled);
		pthread_mutex_unlock(&msg_mutex);

	} // while(1)

}

//------------------------------------------------------------

// Loops on recv() till len bytes have been read
// 's' is an established TCP socket
// 'buffer' is filled with up to 'len' data read
// 'flags' is used for recv()
static inline int sure_recv(const int s, void *buffer, const size_t len, const int flags) {
	int bytesread;
	int offset = 0;
	int toread = len;

	while (toread > 0) {
		bytesread = recv(s, (char*)buffer + offset, toread, flags);
		if (-1 == bytesread) {
			// errno is set by recv
			return -1;
		} else if (0 == bytesread) {
			Dprintf(L_WARNING, "socket closed by remote end (0 bytes read)");
			return 0;
		} else {
			// we read some bytes (possibly not all)
			offset += bytesread;
			toread -= bytesread;
		}
	}
	return offset;
}



// Handles the TCP socket: grabs TCP packets and store them in msg_array[]
// The read is synchronous and will happen only if some free slots in msg_array
// have been found
static void* TCP_thr(void *da_sock) {
	int i;
	int slavesock = (int) da_sock;		// slave socket
	int bytesread;

	Dprintf(D_IOD, "TCP_thr(pid=%d): started as uid=%d euid=%d gid=%d egid=%d",
	     getpid(), getuid(), geteuid(), getgid(), getegid());

#if HAS_PRIO
	if (-1 == setpriority(PRIO_PROCESS, getpid(), PRIO_TCP_THR)) {
		Dprintf(L_FATAL, "setpriority(TCP_thr,%d) failed errno=%d", PRIO_TCP_THR, errno);
		exit(1);
	}
	if (-1 == setregid(user->pw_gid, user->pw_gid)) {
		Dprintf(L_FATAL, "cannot change to %s uid/gid", user->pw_name);
		exit(1);
	}
	
	Dprintf(D_IOD, "TCP_thr(pid=%d): running as uid=%d euid=%d gid=%d egid=%d",
	     getpid(), getuid(), geteuid(), getgid(), getegid());
#endif /* HAS_PRIO */

	// Disable Nagle's algorithm as we only send very small messages as ack's
	do {
		int opt, optlen;
		opt = 1;
		optlen = sizeof(opt);
		if (-1 == setsockopt(slavesock, SOL_TCP, TCP_NODELAY, &opt, optlen)) {
			Dprintf(L_WARNING, "TCP_thr: setsockopt(TCP_NODELAY) failed errno=%d", errno);
		}
	} while(0);


	while (1) {
		// We wait for a slot being freed, event which will eventually
		// happen since there are IO threads that free slots
		pthread_mutex_lock(&msg_mutex);
		while (msg_free == 0) {
			pthread_cond_wait(&msg_cond_free, &msg_mutex);
		}

		// If we're here then a slot has been freed

		// bounded FIFO
		i = iIndexFREE;

		// The following loop cannot be infinite as there's a FREE slot
		while (msg_state[i] != FREE) {
			i = (i + 1) % MSG_ARRAY_SIZE;
		}

		iIndexFREE = (iIndexFREE + 1) % MSG_ARRAY_SIZE;

		// We were signalled there was at least one free slot
		assert(FREE == msg_state[i]);

		// Found a free block so we'll use it
		msg_state[i] = RECV;
		msg_free--;

		// We no longer need the mutex
		pthread_mutex_unlock(&msg_mutex);

		Dprintf(D_IOD, "slot %d: FREE=>RECV [TCP_thr]", i);

		/************************************************************************/

		// Loop till we read a valid message

		Dprintf(D_IOD, "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -");
		bytesread = sure_recv(slavesock, msg_array + i, MSG_IOD_PURGE_LEN, MSG_NOSIGNAL);

		// Only messages of MSG_IOD_PURGE_LEN bytes with a correct magic number are accepted
		if (bytesread != MSG_IOD_PURGE_LEN || msg_array[i].magic != NFSP_MAGIC_MSG) {
			// Do some simplistic checks ...
			// ... about the message size
			if (bytesread < 0) {
				Dprintf(L_WARNING, "recvfrom errno=%d", errno);
			} else if (0 == bytesread) {
				Dprintf(L_WARNING, "socket has been closed");
			} else {
				if (msg_array[i].magic != NFSP_MAGIC_MSG) {
					Dprintf(L_WARNING, "invalid request (not enough magic: 0x%04x instead of 0x%04x)",
							msg_array[i].magic, NFSP_MAGIC_MSG);
				} else {
					Dprintf(L_WARNING, "TCP_thr: internal inconsistency req_size=%d instead of %d",
	    						bytesread, MSG_IOD_PURGE_LEN);
				}
			}
			goto connclosed;
		}

		// We have to send a application-level acknowledgement
		// It's currently a 1 byte message 'X'
		do {
			int ret;
			char buf='X';

			ret = send(slavesock, &buf, 1, MSG_NOSIGNAL);
			if (ret == 1) {
				Dprintf(D_IOD, "ACK sent");
				break;
			}

			if (ret < 0) {
				Dprintf(L_WARNING, "send ack errno=%d", errno);
			} else if (ret == 0) {
				Dprintf(L_WARNING, "closed by peer");
			} else {
				Dprintf(L_WARNING, "TCP_thr: internal inconsistency unexpected ret=%d", ret);
			}
			goto connclosed;
		} while(0);

		/************************************************************************/

		// We need to inform the other threads that there is some work waiting
		pthread_mutex_lock(&msg_mutex);
		msg_state[i] = FILLED;
		msg_filled++;
		assert(0 < msg_filled);
		assert(msg_filled <= MSG_ARRAY_SIZE);

		// Wake up an I/O thread
		pthread_cond_signal(&msg_cond_filled);
		pthread_mutex_unlock(&msg_mutex);

	} // while(1)

	return 0;

	
connclosed:
	// something bad happened so we'd better exit and release the buffer we may still hold
	Dprintf(D_IOD, "closing slavesock");

	if (-1 == close(slavesock)) {
		Dprintf(L_WARNING, "cannot close(slavesock) errno=%d", errno);
	}

	pthread_mutex_lock(&msg_mutex);
	assert(msg_state[i] == RECV);

	msg_state[i] = FREE;
	msg_free++;

	pthread_mutex_unlock(&msg_mutex);

	return 0;
}


//------------------------------------------------------------

static void usage(FILE* const f) {
		fprintf(f,
"Usage: " PROGNAME " [options] datapath\n"
"\n"
"Launches a nfsp IO daemon (iod) chdir()'ed in datapath\n"
"\n"
" -h, --help          this page\n"
" -F, --foreground    don't go background           (default: go background)\n"
" -n x, --nservers x  spawns x I/O servers                      (default: 2)\n"
" -P x, --port x      listens on UDP:x _and_ TCP:x          (default: 20000)\n"
#if HAS_ROOT
" -u x, --user x      switch to uid/gid of user x          (default: nobody)\n"
" -m x, --mode x      choose UDP return mode                                \n"
"                     0:ARP 1:IP 2:UDP                          (default: 0)\n"
#else /* HAS_ROOT */
"\n"
" Send mode is set to UDP\n"
#endif /* !HAS_ROOT */
"\n"
		);
		exit( (f == stderr) ? 1 : 0);
}


void parse_command_line(int argc, char *argv[]) {
	static const char * shortopts = "hFn:P:vu:m:";
	static const struct option longopts[] = {
		{ "help",	0,			0, 'h' },
		{ "foreground",	0,			0, 'F' },
		{ "nservers",	required_argument,	0, 'n' },
		{ "port",	required_argument,	0, 'P' },
#if HAS_ROOT
		{ "user",	required_argument,	0, 'u' },
		{ "sendmode",	required_argument,	0, 'm' },
#endif /* HAS_ROOT */
		{ NULL,		0,			0, 0 },
	};
	char c;

	Dprintf(D_IOD, "About to parse command line");

	opterr = 0;
	while((c = getopt_long(argc, argv, shortopts, longopts, NULL)) != EOF) {
		switch(c) {
			case 'h':
				usage(stdout);
				break;
			case 'F':
				foreground = 1;
				break;
			case 'n':
				nioservers = atoi(optarg);
				if (nioservers <= 0) {
					fprintf(stderr, PROGNAME ": there must be at least 1 server: %s\n", optarg);
					exit(1);
				}
				break;
			case 'P':
				listenport = atoi(optarg);
				if (listenport <= 0) {
					fprintf(stderr, PROGNAME ": port must not be negative: %s\n", optarg);
					exit(1);
				}
				break;
			case 'u':
#if HAS_ROOT
				user = getpwnam(optarg);
				if (NULL == user) {
					fprintf(stderr, PROGNAME ": cannot find user %s\n", optarg);
					exit(1);
				}
#else /* HAS_ROOT */
				Dprintf(L_WARNING, "Warning: user option is not available - HAS_ROOT was not set at compilation time");
#endif /* !HAS_ROOT */
				break;
			case 'm':
#if HAS_ROOT
				xmitmode = atoi(optarg);
				switch (xmitmode) {
					case XMITMODE_ARP:
					case XMITMODE_IP:
					case XMITMODE_UDP:
						break;
					default:
						Dprintf(L_FATAL, "Internal error: unknown xmitmode %d", xmitmode);
						fprintf(stderr, PROGNAME ": unknown xmit mode %d\n", xmitmode);
						exit(1);
				}
#else /* HAS_ROOT */
				Dprintf(L_WARNING, "Warning: sendmode set to UDP - HAS_ROOT was not set at compilation time");
#endif /* !HAS_ROOT */
				break;
			default:
				usage(stderr);
		}
	}

	if (argc - optind == 1) {
		datapath = argv[optind];
	} else {
		fprintf(stderr, PROGNAME ": incorrect number of arguments\n");
		exit(1);
	}
}


int main(int argc, char *argv[]) {
	struct sockaddr_in server_in;
	int i;
	pthread_t thr;

	Dprintf(D_IOD, "Compilation : " __DATE__ " at " __TIME__ " with " __VERSION__ "");

	Dprintf(D_IOD, "main(pid=%d): running as uid=%d euid=%d gid=%d egid=%d",
	     getpid(), getuid(), geteuid(), getgid(), getegid());

	// Check this function since it is parsed as we may still be r00t
	parse_command_line(argc, argv);

#if HAS_ROOT
	if (geteuid() != 0) {
		fprintf(stderr, "Error: this program MUST be launched by root to create low-level sockets.\n");
		exit(1);
	}

//----------------------------------------------------------
	// Get a link layer socket
	llsocket = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_ALL));
	if (llsocket == -1) {
		Dprintf(L_FATAL, "cannot open a link layer socket errno=%d", errno);
		exit(1);
	}

	// Get an IP raw socket
	iprawsocket = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (iprawsocket == -1) {
		Dprintf(L_FATAL, "cannot open a raw socket errno=%d", errno);
		exit(1);
	}
//----------------------------------------------------------

	// Drop the root privileges
	if (NULL == user) {
		user = getpwnam("nobody");
	}
	if (-1 == setegid(user->pw_gid)) {
		Dprintf(L_FATAL, "cannot change to %s's gid=%d errno=%d", user->pw_name, user->pw_gid, errno);
		exit(1);
	}
	if (-1 == seteuid(user->pw_uid)) {
		Dprintf(L_FATAL, "cannot change to %s's uid=%d errno=%d", user->pw_name, user->pw_uid, errno);
		exit(1);
	}
#endif /* HAS_ROOT */

	Dprintf(D_IOD, "main(pid=%d): running as uid=%d euid=%d gid=%d egid=%d",
	     getpid(), getuid(), geteuid(), getgid(), getegid());

#if 0
	// Tell the kernel we'll write most of the IP header by ourselves
	// IP_HDRINCL is enabled by default if proto == IPPROTO_RAW (man 7 raw)
	// To force IP_HDRINCL, use IPPROTO_IP
	do {
		int opt = 1;
		int optlen = sizeof(opt);
		if (- 1 == setsockopt(iprawsocket, SOL_IP, IP_HDRINCL, &opt, optlen)) {
			Dprintf(L_WARNING, "setsockopt(IP_HDRINCL) failed errno=%d", errno);
			exit(1);
		}
	} while (0);
#endif


	// Chdir to the main dir
	if (-1 == chdir(datapath)) {
		Dprintf(L_FATAL, "cannot chdir() to %s", datapath);
		exit(1);
	}

	// Check if subdirs exist and are accessible
	for(i = 0; i < NB_IODS_SUBDIR; i++) {
		char curdir[KEYMAXSIZE];
		sprintf(curdir, "%2.2x", i);

		if (0 == mkdir(curdir, 0700)) {
			Dprintf(L_WARNING, "Warning: created dir: %s\n", curdir);
		}
		strcat(curdir, "/.");
		if (-1 == access(curdir, R_OK|W_OK|X_OK|F_OK)) {
			Dprintf(L_FATAL, "Problem while accessing directory %s errno=%d", curdir, errno);
			exit(1);
		}
	}

//----------------------------------------------------------
	// Mark all message slots as FREE
	for (i = 0; i < MSG_ARRAY_SIZE; i++) {
		msg_state[i] = FREE;
	}

//----------------------------------------------------------
	// Setup the server socket on local port UDPPORT
	memset((char*) &server_in, '\0', sizeof(server_in));
	server_in.sin_family = AF_INET;
	server_in.sin_addr.s_addr = INADDR_ANY;
	server_in.sin_port = htons(listenport);

	udpsocket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (udpsocket == -1) {
		Dprintf(L_FATAL, "cannot open a UDP socket errno=%d", errno);
		exit(1);
	}
	if (-1 == bind(udpsocket, (struct sockaddr *) &server_in, sizeof(server_in))) {
		Dprintf(L_FATAL, "cannot bind UDP socket %d errno=%d", udpsocket, errno);
		exit(1);
	}

//----------------------------------------------------------
	forward_step1_read_udpsock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (forward_step1_read_udpsock == -1) {
		Dprintf(L_FATAL, "cannot open a UDP socket (forward_step1_read_udpsock) errno=%d", errno);
		exit (1);
	}

	forward_step1_write_udpsock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (forward_step1_write_udpsock == -1) {
		Dprintf(L_FATAL, "cannot open a UDP socket (forward_step1_write_udpsock) errno=%d", errno);
		exit (1);
	}

#ifdef WRITEMIRROR_SUPPORT
	write_mirror_udpsock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (write_mirror_udpsock == -1) {
		Dprintf(L_FATAL, "cannot open a UDP socket (write_mirror_udpsock) errno=%d", errno);
		exit (1);
	}
#endif /* WRITEMIRROR_SUPPORT */

//----------------------------------------------------------
	// Setup the server socket on local port TCPORT
	memset((char*) &server_in, '\0', sizeof(server_in));
	server_in.sin_family = AF_INET;
	server_in.sin_addr.s_addr = INADDR_ANY;
	server_in.sin_port = htons(listenport);

	tcpsocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (-1 == tcpsocket) {
		Dprintf(L_FATAL, "main: cannot open a TCP socket errno=%d", errno);
		exit(1);
	}

	do {
		int opt = 1;
		int optlen = sizeof(opt);
		if (-1 == setsockopt(tcpsocket, SOL_SOCKET, SO_REUSEADDR, &opt, optlen)) {
			Dprintf(L_FATAL, "main: cannot setsockopt(SO_REUSEADDR) errno=%d", errno);
			exit(1);
		}
	} while (0);

	if (-1 == bind(tcpsocket, (struct sockaddr *) &server_in, sizeof(server_in))) {
		Dprintf(L_FATAL, "main: cannot bind() TCP socket errno=%d", errno);
		exit(1);
	}

	if (-1 == listen(tcpsocket, TCP_BACKLOG)) {
		Dprintf(L_FATAL, "main: cannot listen() on TCP socket errno=%d", errno);
		exit(1);
	}


//----------------------------------------------------------
	udpspoof_init();

	enable_logging("iod");

//----------------------------------------------------------
	if (!foreground) {
		// user wants a daemon
		pid_t pid=fork();
		switch (pid) {
			case -1:		// error
				Dprintf(L_FATAL, "cannot daemonize (fork failed) errno=%d", errno);
				exit(1);
				break;
			case 0:			// fork()'ed process
			// FIXME: Clean up logging
			do {
				int ret;
				int nullfd;
				
				nullfd = open("/dev/null", O_RDWR);
				if (nullfd < 0) {
					Dprintf(L_FATAL, "cannot open /dev/null for read/write");
					exit(1); 
				}
				
				/* Now we remove ourselves from the foreground. */
				ret = close(0);
				ret = close(1) || ret;
				ret = close(2) || ret;
				if (ret) {
					syslog(LOG_INFO, "cannot close fd 0/1/2 - exiting");
					exit(1);
				}

				ret = (0 != dup2(nullfd, 0));
				ret = (1 != dup2(nullfd, 1)) || ret;
				ret = (2 != dup2(nullfd, 2)) || ret;
				ret = close(nullfd) || ret;
				if (ret) {
       					syslog(LOG_INFO, "cannot dup2 nullfd on 0/1/2 - exiting");
					exit(1);
				}
//#ifdef HAVE_SETSID
				setsid();
//#endif /* HAVE_SETSID */
			} while(0);
				break;
			default:		// original process
				exit(0);
		}
	}

//----------------------------------------------------------
	srand(getpid());
	set_ipid(rand());

//----------------------------------------------------------
	for(i = 0; i < nioservers; i++) {
		if (pthread_create(&thr, NULL, IO_thr, NULL)) {
			Dprintf(L_FATAL, "main: pthread_create(IO_thr) failed");
			exit(1);
		}
		if (pthread_detach(thr)) {
			Dprintf(L_FATAL, "main: cannot pthread_detach(IO_thr)");
			exit(1);
		}
	}

	if (pthread_create(&thr, NULL, UDP_thr, NULL)) {
		Dprintf(L_FATAL, "main: cannot pthread_create(UDP_thr)");
		exit(1);
	}
	if (pthread_detach(thr)) {
		Dprintf(L_FATAL, "main: cannot pthread_detach(UDP_thr)");
		exit(1);
	}

	// We are executing here as user's euid/egid but our uid/gid is still 0
	while (1) {
		struct sockaddr_in sin;
		socklen_t sin_len;
		int sock;

		sin_len = sizeof(sin);
		sock = accept(tcpsocket, (struct sockaddr *) &sin, &sin_len);
		if (sock == -1) {
			Dprintf(L_WARNING, "accept failed on tcpsocket=%d errno=%d", tcpsocket, errno);
			continue;
		}
		Dprintf(D_IOD, "Accepted connection from " NTODSTR "(TCP)", NTOD(sin));

		if (pthread_create(&thr, NULL, TCP_thr, (void *) sock)) {
			Dprintf(L_FATAL, "main: cannot pthread_create(TCP_thr)");
			exit(1);
		}
		if (pthread_detach(thr)) {
			Dprintf(L_FATAL, "main: cannot pthread_detach(TCP_thr)");
			exit(1);
		}

		Dprintf(D_IOD, "TCP thread %ld up and running", thr);
	}

	return 0;
}

