// unlinkd.c
//
// This file implements the purge daemon:
//   * gets requests on a pipe from the nfsd
//   * send them :
//     - when the message is completely filled
//     - when some time has elapsed without receiving nothing
//       (constant UNLINKD_TIMEOUT)

#include "default.h"

#ifdef VIODS_SUPPORT
#include "nfsp_viods.h"
#else /* VIODS_SUPPORT */
#include "nfsp_iods.h"
#endif /* !VIODS_SUPPORT */

#include "nfsd.h"
#include "iodng.h"
#include "signals.h"

//#define SHOWINFO 1
//
#include "ipstring.h"
#include "logging.h"
//#include "dbg.h"

#include "unlinkd.h"

#include <sys/poll.h>
#include <syslog.h>


#ifndef VIODS_SUPPORT
// The array of sockets used to talk to iods (TCP)
static int iods_tcp_fds[MAX_IODS];

// The message used to send purge requests to iod's
static msg_iod_t rm_msg;

// Connects (synchronous) for iod #i
// Updates iods_tcp_fds[i]
static int reconnect(int i) {
	if (-1 != iods_tcp_fds[i]) {
		if (-1 == close(iods_tcp_fds[i])) {
			Dprintf(L_WARNING, "close () failed errno=%d", errno);
		}
	}

	iods_tcp_fds[i] = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (-1 == iods_tcp_fds[i]) {
		Dprintf(L_WARNING, "socket() failed errno=%d", errno);
		return -1;
	}

	// It's a blocking connect under the assumption it will happen not
	// often in standard mode
	if(-1 == connect(iods_tcp_fds[i], iods+i, sizeof(struct sockaddr_in))) {
		Dprintf(L_WARNING, "cannot connect to iod=%d " NTODSTR " errno=%d",
			i, NTOD(iods[i]), errno);
		return -1;
	}

	Dprintf(D_UNLINKD, "connection to iod=%d " NTODSTR " successful", i, NTOD(iods[i]));
	return 0;
}


static void multicast_remove() {
	int i;
	int ret;
	
	Dprintf(D_UNLINKD, "multicast_remove() starts");

	if (0 == rm_msg.purgereq.nfiles) {
		Dprintf(D_UNLINKD, "no message to send => returning");
		return;
	}

	// We mark the remaining slots in the message as "non files" by
	// setting their inodes number to 0 (invalid inode)
	for(i = rm_msg.purgereq.nfiles; i < MAXPENDINGFILES; ++i) {
	        rm_msg.purgereq.files[i].inode = 0;
	}

	for(i = 0; i < iods_nb; ++i) {
tryagain:
		ret = send(iods_tcp_fds[i], (char *) &rm_msg, MSG_IOD_PURGE_LEN, MSG_NOSIGNAL);
		if (MSG_IOD_PURGE_LEN != ret) {
			if (ret == -1) {
				if (EPIPE == errno) {
					Dprintf(L_WARNING, "multicast_remove send() failed (EPIPE) reconnecting iod %d", i);
					if (-1 == reconnect(i)) {
						// We spin on reconnect since if the iod is down all the
						// system will be soon down...
						Dprintf(L_WARNING, "reconnect failed ... sleeping %d ms before retrying", UNLINKD_RETRYCONNECT);
						usleep(1000 * UNLINKD_RETRYCONNECT);
					} else {
						Dprintf(L_WARNING, "reconnect succeeded... retrying");
					}
					goto tryagain;
				} else {
					Dprintf(L_WARNING, "multicast_remove send() for iod %d failed errno=%d", 
						i, errno);
				}
			} else {
				Dprintf(L_WARNING, "multicast_remove send() for iod %d sent %d bytes instead of %d", 
					i, ret, MSG_IOD_PURGE_LEN);
				//XXX
			}
		} else {
			char c;
			ret = recv(iods_tcp_fds[i], &c, 1, MSG_NOSIGNAL);
			switch(ret) {
				case -1:
				case 0:
					if (-1 == reconnect(i)) {
						Dprintf(L_WARNING, "reconnect failed ... sleeping %d ms before retrying", UNLINKD_RETRYCONNECT);
						usleep(1000 * UNLINKD_RETRYCONNECT);
					} else {
						Dprintf(L_WARNING, "reconnect succeeded... retrying");
					}
					goto tryagain;
					break;
				case 1:
					if (c != 'X') {
						if (-1 == reconnect(i)) {
							Dprintf(L_WARNING, "reconnect failed ... sleeping %d ms before retrying", UNLINKD_RETRYCONNECT);
							usleep(1000 * UNLINKD_RETRYCONNECT);
						} else {
							Dprintf(L_WARNING, "reconnect succeeded... retrying");
						}
						goto tryagain;
					}

					// everything is OK
					Dprintf(D_UNLINKD, "msg sent and ack'ed by iod %d successfully", i);
					// bingo !!!
					
			}
			
		}
	}

	rm_msg.purgereq.nfiles=0;
}
#endif /* !VIODS_SUPPORT */

void go_background() {
	int ret;
	int nullfd;

	nullfd = open("/dev/null", O_RDWR);
	if (nullfd < 0) {
		Dprintf(L_FATAL, "cannot open /dev/null for read/write");
		exit(1);
	}

	/* Now we remove ourselves from the foreground. */
	ret = close(0);
	ret = close(1) || ret;
	ret = close(2) || ret;

	if (ret) {
		Dprintf(L_FATAL, "cannot close fd 0/1/2 - exiting");
		syslog(LOG_INFO, "cannot close fd 0/1/2 - exiting");
		exit(1);
	}

	ret = (0 != dup2(nullfd, 0));
	ret = (1 != dup2(nullfd, 1)) || ret;
	ret = (2 != dup2(nullfd, 2)) || ret;
	ret = close(nullfd) || ret;
	if (ret) {
		Dprintf(L_FATAL, "cannot dup2 nullfd on 0/1/2 - exiting");
		syslog(LOG_INFO, "cannot dup2 nullfd on 0/1/2 - exiting");
		exit(1);
	}

#ifdef HAVE_SETSID
		setsid();
#else
	do {
		int fd;

		if ((fd = open("/dev/tty", O_RDWR)) >= 0) {
			ioctl(fd, TIOCNOTTY, (char *) NULL);
			close(fd);
		}
	} while (0);
#endif
}




void unlinkd(int foreground) {
	remove_req_t rm_req;
	int ret;
#ifndef VIODS_SUPPORT
	struct pollfd pollpipe;
	int i;
#endif /* !VIODS_SUPPORT */

	if (!foreground) {
		go_background();
	}
	Dprintf(D_UNLINKD, "unlink daemon started pipe=%d", unlinkpipe);

	// We remove signal handlers that have been set by our father
	ignore_signal(SIGHUP);
	ignore_signal(SIGCHLD);
	ignore_signal(SIGQUIT);
	ignore_signal(SIGIOT);
	ignore_signal(SIGUSR1);
	ignore_signal(SIGUSR2);
	ignore_signal(SIGALRM);
	ignore_signal(SIGPIPE);


#ifdef VIODS_SUPPORT
#warning "VIODS_SUPPORT for unlinkd to write"
	Dprintf(L_WARNING,"unlinkd for viod not implemented => requests dropped");
	while (1) {
		ret = read(unlinkpipe, &rm_req, sizeof(rm_req));
	}
#else /* VIODS_SUPPORT */

	// Initializations
	for(i=0;i<MAX_IODS;++i) {
		iods_tcp_fds[i]=-1;
	}

	Dprintf(D_UNLINKD, "About to connect to the %d iods", iods_nb);
	ret=0;
	for(i=0; i < iods_nb; ++i) {
		if (-1 == reconnect(i)) {
			++ret;
		}
	}
	if (ret) {
		Dprintf(L_FATAL, "aborting: cannot establish connection to some iods");
		exit(1);
	}
	Dprintf(D_UNLINKD, "All the connections look OK");

	Dprintf(D_UNLINKD, "init_iods(): initializing the PURGE msg");
	rm_msg.magic = NFSP_MAGIC_MSG;
	rm_msg.msgtype = NFSP_PURGE_REQ;
	rm_msg.purgereq.nfiles = 0;

	while (1) {
		pollpipe.fd = unlinkpipe;
		pollpipe.events = POLLIN;
		pollpipe.revents = 0;

		ret = poll(&pollpipe, 1, UNLINKD_TIMEOUT);


		switch(ret) {
			case -1:
				Dprintf(L_WARNING, "poll did not succeed errno=%d",errno);
				continue;
				break;

			case 0:
				// time out => we do not wait any longer and send the requests to the iods
				multicast_remove();
				continue;
				break;
				
			case 1:
				if (! (pollpipe.revents & POLLIN)) {
					continue;
				}

				ret = read(unlinkpipe, &rm_req, sizeof(rm_req));
				if (ret != sizeof(rm_req)) {
					switch (ret) {
						case -1:
							Dprintf(L_WARNING, "poll cannot read in pipe %d errno=%d",
									unlinkpipe, errno);
							break;

						case 0:
							// if 0 byte is read then it's EOF
							Dprintf(L_WARNING, "read 0 byte in pipe %d => exit()", unlinkpipe);
							break;

						default:
							Dprintf(L_WARNING, "short read in pipe %d: %d bytes instead of %d",
								unlinkpipe, ret, sizeof(remove_req_t));
					}
					exit(1);

				} else {
					if (REMOVE_REQ_MAGIC != rm_req.magic) {
						Dprintf(L_WARNING, "invalid magic number got=%x whereas expected=%x",
							rm_req.magic, REMOVE_REQ_MAGIC);
						continue;
					}

					rm_msg.purgereq.files[rm_msg.purgereq.nfiles].inode = rm_req.inode;
					rm_msg.purgereq.files[rm_msg.purgereq.nfiles].seed = rm_req.seed;
					++rm_msg.purgereq.nfiles;
	
					if (rm_msg.purgereq.nfiles >= MAXPENDINGFILES) {
						// Time to send the message
						multicast_remove();
					}
					break;
				}

			default:
				Dprintf(L_WARNING, "poll() smoked crack and returned ret=%d", ret);
		}

	} // end of while(1)
#endif /* !VIODS_SUPPORT */

	while(1) {
		sleep(UINT_MAX);
	}
}

